use rand::Rng;
use serde::{Deserialize, Serialize};

use crate::fraction::Fraction;

#[derive(Debug, Copy, Clone, Serialize, Deserialize)]
pub enum FrequencyShifting {
    Linear,
    Quadratic,
    Cubic,
    Ease,
    Pulsating,
    Spiking,
    Randomize,
    Maniac,
    Synchronized,
}

impl TryFrom<u32> for FrequencyShifting {
    type Error = String;

    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(FrequencyShifting::Linear),
            1 => Ok(FrequencyShifting::Quadratic),
            2 => Ok(FrequencyShifting::Cubic),
            3 => Ok(FrequencyShifting::Ease),
            4 => Ok(FrequencyShifting::Pulsating),
            5 => Ok(FrequencyShifting::Spiking),
            6 => Ok(FrequencyShifting::Randomize),
            7 => Ok(FrequencyShifting::Maniac),
            8 => Ok(FrequencyShifting::Synchronized),
            _ => Err(String::from("Unregconized shifting mode")),
        }
    }
}

pub fn deserialize_frequency_shifting<'de, D>(
    deserializer: D,
) -> Result<FrequencyShifting, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let value = u32::deserialize(deserializer)?;
    FrequencyShifting::try_from(value).map_err(serde::de::Error::custom)
}

pub fn shifting_frequency(
    sample_tick: u32,
    start_timing: u32,
    start_pulse: &crate::pattern::Pulse,
    end_pulse: &crate::pattern::Pulse,
) -> (u16, u16, u8) {
    match end_pulse.frequency_shifting {
        FrequencyShifting::Linear => bezier_shifting(
            sample_tick,
            start_timing,
            start_pulse.frequency,
            start_timing + end_pulse.offset,
            end_pulse.frequency,
            (0.0, 0.0),
            (1.0, 1.0),
        ),
        FrequencyShifting::Quadratic => bezier_shifting(
            sample_tick,
            start_timing,
            start_pulse.frequency,
            start_timing + end_pulse.offset,
            end_pulse.frequency,
            (1.0, 0.0),
            (1.0, 1.0),
        ),
        FrequencyShifting::Cubic => bezier_shifting(
            sample_tick,
            start_timing,
            start_pulse.frequency,
            start_timing + end_pulse.offset,
            end_pulse.frequency,
            (0.9, 0.1),
            (1.0, 0.2),
        ),
        FrequencyShifting::Ease => bezier_shifting(
            sample_tick,
            start_timing,
            start_pulse.frequency,
            start_timing + end_pulse.offset,
            end_pulse.frequency,
            (0.33, 0.0),
            (0.67, 1.0),
        ),
        FrequencyShifting::Pulsating => periodic_shifting(
            sample_tick,
            start_timing,
            start_timing + end_pulse.offset,
            start_pulse.frequency.min(end_pulse.frequency),
            end_pulse.frequency.max(start_pulse.frequency),
        ),
        FrequencyShifting::Spiking => periodic_shifting(
            sample_tick,
            start_timing,
            start_timing + end_pulse.offset,
            start_pulse.frequency.min(end_pulse.frequency),
            start_pulse.frequency.max(end_pulse.frequency) * 2.0,
        ),
        FrequencyShifting::Randomize => randomize_shifting(
            start_pulse.frequency.min(end_pulse.frequency),
            start_pulse.frequency.max(end_pulse.frequency),
        ),
        FrequencyShifting::Maniac => randomize_shifting(1.0, 1000.0),
        FrequencyShifting::Synchronized => bezier_shifting(
            sample_tick,
            start_timing,
            start_pulse.frequency,
            start_timing + end_pulse.offset,
            end_pulse.frequency,
            end_pulse.control_point_1.into(),
            end_pulse.control_point_2.into(),
        ),
    }
}

fn estimate_frequency_level(frequency: f64) -> u8 {
    match frequency {
        0.0..=150.0 => 1,
        150.0..=300.0 => 2,
        300.0..=600.0 => 3,
        600.0..=800.0 => 4,
        800.0.. => 5,
        _ => 0,
    }
}

fn resolve_frequency(frequency: f64) -> (u16, u16) {
    let waveform_frequency = frequency * 1000_f64;
    let x = (waveform_frequency.sqrt() * 15_f64).floor() as i64;
    let y = (waveform_frequency - x as f64).floor() as i64;
    let f = Fraction::new(x, y);
    let output = u16::try_from(f.numerator).unwrap_or(u16::MAX);
    let idle = u16::try_from(f.denominator).unwrap_or(u16::MAX);
    (output, idle)
}

fn bezier_shifting(
    sample_tick: u32,
    start_time: u32,
    start_frequency: f64,
    end_time: u32,
    end_frequency: f64,
    control_1: (f64, f64),
    control_2: (f64, f64),
) -> (u16, u16, u8) {
    let t = (sample_tick - start_time) as f64 / (end_time - start_time) as f64;
    let u = 1.0 - t;
    let tt = t * t;
    let uu = u * u;
    let uuu = uu * u;
    let ttt = tt * t;

    let p0 = (start_time as f64, start_frequency);
    let p1 = control_1;
    let p2 = control_2;
    let p3 = (end_time as f64, end_frequency);

    let frequency = uuu * p0.1 + 3.0 * uu * t * p1.1 + 3.0 * u * tt * p2.1 + ttt * p3.1;

    let (x, y) = resolve_frequency(frequency);
    let estimate_level = estimate_frequency_level(frequency);
    (x, y, estimate_level)
}

fn periodic_shifting(
    sample_tick: u32,
    start_time: u32,
    _end_time: u32,
    min_frequency: f64,
    max_frequency: f64,
) -> (u16, u16, u8) {
    let period_length = {
        let mut l = sample_tick - start_time;
        while l > 25 {
            l -= 25;
        }
        l
    };
    let frequency = if (sample_tick - start_time) / period_length % 2 == 0 {
        max_frequency
    } else {
        min_frequency
    };
    let (x, y) = resolve_frequency(frequency);
    let estimate_level = estimate_frequency_level(frequency);
    (x, y, estimate_level)
}

fn randomize_shifting(min_frequency: f64, max_frequency: f64) -> (u16, u16, u8) {
    let mut rng = rand::thread_rng();
    let frequency = rng.gen_range(min_frequency..=max_frequency);
    let (x, y) = resolve_frequency(frequency);
    let estimate_level = estimate_frequency_level(frequency);
    (x, y, estimate_level)
}
