use std::collections::VecDeque;

use serde::{Deserialize, Serialize};
use uuid::Uuid;

use super::{
    frequency_shifting::{self, deserialize_frequency_shifting, FrequencyShifting},
    SMOOTH_REPEAT_FADE_OFFSET,
};
use crate::fraction::Fraction;

pub fn calculate_output_ratio(waveform_frequency: f64) -> Fraction {
    let x = waveform_frequency.sqrt() * 15f64;
    let y = waveform_frequency * 1000f64 - x;
    Fraction::new_from_float(x, y)
}

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
pub struct CanvasPoint {
    pub x: f64,
    pub y: f64,
}

impl Into<(f64, f64)> for CanvasPoint {
    fn into(self) -> (f64, f64) {
        (self.x, self.y)
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Pulse {
    pub order: u32,
    pub id: Uuid,
    #[serde(default)]
    pub offset: u32,
    pub width: u32,
    pub maniac: bool,
    pub frequency: f64,
    #[serde(deserialize_with = "deserialize_frequency_shifting")]
    pub frequency_shifting: FrequencyShifting,
    pub control_point_1: CanvasPoint,
    pub control_point_2: CanvasPoint,
}

#[derive(Debug, Clone, Copy, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct RawPulse {
    pub tick_order: u32,
    pub z: u16,
    pub x: u16,
    pub y: u16,
    pub frequency_level: u8,
}

fn bezier(
    t: f64,
    p0: CanvasPoint,
    p1: CanvasPoint,
    cp1: CanvasPoint,
    cp2: CanvasPoint,
) -> CanvasPoint {
    let mut p = CanvasPoint { x: t, y: 0.0 };

    let t = (t - p0.x) / (p1.x - p0.x);
    let u = 1.0 - t; // 1-t
    let tt = t * t; // t^2
    let uu = u * u; // (1-t)^2
    let uuu = uu * u; // (1-t)^3
    let ttt = tt * t; // t^3

    p.y = uuu * p0.y; // (1-t)^3 * p0.y
    p.y += 3.0 * uu * t * cp1.y; // 3 * (1-t)^2 * t * cp1.y
    p.y += 3.0 * u * tt * cp2.y; // 3 * (1-t) * t^2 * cp2.y
    p.y += ttt * p1.y; // t^3 * p1.y

    p
}

pub fn generate_pulse_sequence<P: IntoIterator<Item = Pulse>>(
    pulse: P,
    smooth_repeat: bool,
) -> Box<dyn Iterator<Item = RawPulse>> {
    let sorted_pulses = pulse.into_iter().collect::<VecDeque<_>>();

    let mut sample_tick = 0;
    let mut timing = 0;
    let mut sample_queue: VecDeque<(u32, u32, Pulse, Pulse)> = VecDeque::new();
    let head_pulse = sorted_pulses.front();
    let mut peekable_pulses = sorted_pulses.iter().peekable();
    while let Some(p) = peekable_pulses.next() {
        if let Some(next_pulse) = peekable_pulses.peek() {
            while sample_tick < (timing + next_pulse.offset) {
                sample_queue.push_back((sample_tick, timing, p.clone(), (*next_pulse).clone()));
                sample_tick += 25;
            }
            timing += next_pulse.offset;
        } else if smooth_repeat && head_pulse.is_some() {
            sample_queue.push_back((
                sample_tick,
                timing,
                p.clone(),
                Pulse {
                    offset: SMOOTH_REPEAT_FADE_OFFSET,
                    order: p.order + 1,
                    ..head_pulse.unwrap().clone()
                },
            ));
        } else {
            break;
        }
    }

    Box::new(std::iter::from_fn(move || {
        if let Some((tick, timing, prev, next)) = sample_queue.pop_front() {
            let bezier_point = bezier(
                tick as f64,
                CanvasPoint {
                    x: timing as f64,
                    y: prev.frequency,
                },
                CanvasPoint {
                    x: (timing + next.offset) as f64,
                    y: next.frequency,
                },
                prev.control_point_1,
                prev.control_point_2,
            );
            let (x, y, frequency_level) =
                frequency_shifting::shifting_frequency(tick, timing, &prev, &next);
            return Some(RawPulse {
                tick_order: tick,
                z: bezier_point.y as u16,
                x,
                y,
                frequency_level,
            });
        }
        None
    }))
}
