import { Icon } from '@iconify/react/dist/iconify.js';
import cx from 'clsx';
import { useAtom, useAtomValue, useSetAtom } from 'jotai';
import { FC, useCallback, useMemo, useState } from 'react';
import { useNavigate } from 'react-router-dom';
import { useDebounce } from 'react-use';
import { ScrollArea } from '../../components/ScrollArea';
import {
  Pattern,
  PatternsAtom,
  SearchKeywordAtom,
  SelectedPatternIdAtom,
  totalDuration,
} from '../../context/Patterns';
import styles from './Patterns.module.css';

const PatternCard: FC<{ pattern: Pattern }> = ({ pattern }) => {
  const [selectedId, setSelectedId] = useAtom(SelectedPatternIdAtom);
  const navigate = useNavigate();
  const duration = useMemo(() => {
    return totalDuration(pattern) / 1000;
  }, [pattern]);
  const selected = useMemo(() => selectedId === pattern.id, [selectedId, pattern]);
  const handleSingleClick = useCallback(() => {
    if (selectedId === pattern.id) {
      setSelectedId(null);
    } else {
      setSelectedId(pattern.id);
    }
  }, [pattern, selectedId]);
  const handleDblClick = useCallback(() => {
    setSelectedId(pattern.id);
    navigate('/pattern-editor/edit');
  }, [pattern]);

  return (
    <div
      className={cx(styles.pattern_card, selected && styles.selected)}
      onClick={handleSingleClick}
      onDoubleClick={handleDblClick}>
      <h5 className={styles.name}>{pattern.name}</h5>
      <span>{duration.toFixed(2)} s</span>
    </div>
  );
};

const Patterns: FC = () => {
  const [rawKeyword, setRawKeyword] = useState<string>('');
  const [keyword, setKeyword] = useAtom(SearchKeywordAtom);
  const patterns = useAtomValue(PatternsAtom);
  const setPatternSelection = useSetAtom(SelectedPatternIdAtom);
  const navigate = useNavigate();
  const createNewAction = useCallback(() => {
    setPatternSelection(null);
    navigate('/pattern-editor/new');
  }, []);

  useDebounce(
    () => {
      if (rawKeyword.trim().length > 0) {
        setKeyword(rawKeyword.trim());
      } else {
        setKeyword(null);
      }
    },
    1500,
    [rawKeyword],
  );

  return (
    <div className={styles.patterns}>
      <div className={styles.search_row}>
        <div className={cx('input_wrapper', 'extendable')}>
          <Icon icon="material-symbols-light:search" />
          <input
            type="text"
            placeholder="Search patterns..."
            value={rawKeyword}
            onChange={(evt) => setRawKeyword(evt.currentTarget.value)}
          />
        </div>
        <button className="tonal secondary" onClick={createNewAction}>
          <Icon icon="material-symbols-light:add" />
          New Pattern
        </button>
      </div>
      <ScrollArea enableY>
        <div className={styles.pattern_list}>
          {patterns.length === 0 && <div className="empty_prompt">No pattern found.</div>}
          {patterns.map((p) => (
            <PatternCard key={p.id} pattern={p} />
          ))}
        </div>
      </ScrollArea>
    </div>
  );
};

export default Patterns;
